/********************************************
 * UTIA AV CR, v.v.i
 * Signal Processing Department
 * Pod Vodarenskou vezi 4
 * 182 08 Prague
 * Czech Republic
 * web: http://zs.utia.cas.cz
 *
 * Filename: main.c
 * Date: 02.01.2016
 * Authors:
 * Lukas Kohout <kohoutl@utia.cas.cz>
 * Zdenek Pohl <xpohl@utia.cas.cz>
 * Jiri Kadlec <kadlec@utia.cas.cz>
 * Description:
 *
 * Changelog:
 *
 *
 ********************************************/

#include <stdio.h>
#include <stdlib.h>

#include <xil_cache.h>
#include <xparameters.h>
#include <platform.h>

#include "ui.h"
#include "img_filters.h"

// ***************************************************************************
#define I2C_BASE  XPS_I2C0_BASEADDR
#define DDR_MEM_BASE  XPAR_DDR_MEM_BASEADDR
// ***************************************************************************

#include "sds_lib.h"
unsigned long long clock_start, clock_end_processing, clock_end_frame;
#define TIME_STAMP_INIT    clock_start = sds_clock_counter();
#define TIME_STAMP_1  { clock_end_processing = sds_clock_counter();}
#define TIME_STAMP_2  { clock_end_frame = sds_clock_counter();}
#define TIME_STAMP_PRINT {printf("Image processing time: %llu, Total FPS: %f \r\n", clock_end_processing-clock_start, (float)XPAR_CPU_CORTEXA9_0_CPU_CLK_FREQ_HZ/(float)(clock_end_frame-clock_start)); }

void sobel_demo_processing(unsigned short *in_buffer,
		unsigned short *out_buffer) {
	TIME_STAMP_2
	TIME_STAMP_PRINT
	TIME_STAMP_INIT
	img_process(in_buffer, out_buffer);
	TIME_STAMP_1
}

// ***************************************************************************
// Video Timing Controller VTC
#include "vtc.h"

vtc_t vtc;
extern timing_t vtc_resolutions[VTC_RES_CNT];

// ***************************************************************************
// Video DMA VDMA
#include "video_frame_buffer.h"

#define MAX_RES_X       1920
#define MAX_RES_Y       1200 //1080
#define MAX_FRAME_SIZE  2048*8192              // max size of one frame for VDMA aligned to kB

#define DDR_MEM_BASE  XPAR_DDR_MEM_BASEADDR
#define FB_OFFSET       0x20000000  // 512 MB
#define FB_BASE         DDR_MEM_BASE + FB_OFFSET

video_resolution_t fb_res;
video_resolution_t fb_store_res;

/*
 * Frame offset in VDMA
 */
video_offset_t offset = { 0, 0 };

// ***************************************************************************
// IMAGEON
#include <fmc_iic.h>
#include <fmc_imageon.h>
#include "fmc_imageon_iic.h"
#include "fmc_imageon_hdmi_framebuffer.h"

fmc_imageon_hdmi_framebuffer_t fb;

#include <tcm_receiver.h>
#include "tcm5117pl.h"
#define TCM_RECV_BASE XPAR_TCM_RECEIVER_COLOR_TCM_RECEIVER_0_S00_AXI_BASEADDR


// ***************************************************************************

int tcm_init(fmc_imageon_hdmi_framebuffer_t *fb)
{
	u32 status;

	//   tcm5117pl_reset(XPAR_CAM_RESET_BASEADDR);

	/*
	 * switch IMAGEON i2c mux to camera port
	 */
	fmc_imageon_iic_mux( &(fb->fmc_imageon), FMC_IMAGEON_I2C_SELECT_CAM );

	/*
	 * get and print camera ID
	 */
	tcm5117pl_get_chip_id( &(fb->fmc_imageon_iic) );

	/*
	 * TCM core ISERDES Reset Assert
	 */
	TCM_RECEIVER_mWriteReg(TCM_RECV_BASE, 0x00, 1);

	/*
	 * configure the camera - 1920x1080p60
	 */
	status = tcm5117pl_init( &(fb->fmc_imageon_iic), TCM5117PL_1080P60 );

	/*
	 * TCM core ISERDES Reset De-Assert
	 */
	TCM_RECEIVER_mWriteReg(TCM_RECV_BASE, 0x00, 0);

	if ( status != XST_SUCCESS) return XST_FAILURE;

	return XST_SUCCESS;
}



#define advanceFStore(f)  {f = (f + 1) % fb.uNumFrames_HdmiFrameBuffer;}

// ***************************************************************************
// ***************************************************************************

// ***************************************************************************
// define black color in YCrCb 4:2:2 color space
#define PIX_Y_INIT     0x00
#define PIX_CBCR_INIT  0x80

void clear_fb_YCrCb(fmc_imageon_hdmi_framebuffer_t *fb) {
	volatile u8 *pStorageMem;
	int i;

	pStorageMem = (u8 *) fb->uBaseAddr_MEM_HdmiFrameBuffer;
	for (i = 0; i < fb->uNumFrames_HdmiFrameBuffer * MAX_FRAME_SIZE;
			i += fb->vdma_rx_bytes) {
		// Init Pixels
		*pStorageMem++ = PIX_Y_INIT;
		*pStorageMem++ = PIX_CBCR_INIT;
	}
	Xil_DCacheFlushRange(fb->uBaseAddr_MEM_HdmiFrameBuffer,
			fb->uNumFrames_HdmiFrameBuffer * MAX_FRAME_SIZE);
}

// ***************************************************************************
int set_resolution(vtc_t *vtc, fmc_imageon_hdmi_framebuffer_t *fb, u8 sel,
		video_resolution_t *res) {
	int status = 0;
	double dummy;

	/*
	 * set new pixel clock
	 */
	xil_printf("Initialize CDCE913 ... ");
	xil_printf(" %d.%d MHz ... ",
			(unsigned int) my_floor(vtc_resolutions[sel].PClk),
			(unsigned int) (my_modf(vtc_resolutions[sel].PClk, &dummy) * 1000.0));
	status = fmc_imageon_vclk_set(&(fb->fmc_imageon),
			vtc_resolutions[sel].PClk);
	if (status != XST_SUCCESS) {
		xil_printf("Failed\r\n");
		return XST_FAILURE;
	} else {
		xil_printf("OK\r\n");
	}

	/*
	 * Initialize VDMA TX
	 */
	xil_printf("Initialize VDMA ... ");
	status = vfb_tx_stop(&fb->vdma_hdmi);
	if (status != XST_SUCCESS) {
		xil_printf("FAILED - Cannot stop VDMA 1 TX\r\n");
		return XST_FAILURE;
	}

	res->x = vtc_resolutions[sel].HActiveVideo;
	res->y = vtc_resolutions[sel].VActiveVideo;

	xil_printf("TX Init ... ");
	status = vfb_tx_init(&fb->vdma_hdmi,                     // pAxiVdma
			&fb->vdmacfg_hdmi_read,             // pReadCfg
			res,                                // uVideoResolution
			&fb_store_res,       // uStorageResolution  -- always set to maximum
			fb->uBaseAddr_MEM_HdmiFrameBuffer,  // uMemAddr
			fb->uNumFrames_HdmiFrameBuffer,     // uNumFrames
			&offset,                           // horizontal and vertical offset
			fb->vdma_tx_bytes                   // number of bytes per pixel
	);
	if (status != XST_SUCCESS) {
		xil_printf("FAILED\n\r");
		return XST_FAILURE;
	} else {
		xil_printf("OK\n\r");
	}

	/*
	 * initialize IMAGEON HDMIO
	 */
	xil_printf("HDMIO Initialization ... ");
	fb->hdmio_timing.IsHDMI = 0;

	fb->hdmio_timing.HActiveVideo = vtc_resolutions[sel].HActiveVideo;
	fb->hdmio_timing.HFrontPorch = vtc_resolutions[sel].HFrontPorch;
	fb->hdmio_timing.HSyncWidth = vtc_resolutions[sel].HSyncWidth;
	fb->hdmio_timing.HBackPorch = vtc_resolutions[sel].HBackPorch;
	fb->hdmio_timing.HSyncPolarity = vtc_resolutions[sel].HSyncPolarity;

	fb->hdmio_timing.VActiveVideo = vtc_resolutions[sel].VActiveVideo;
	fb->hdmio_timing.VFrontPorch = vtc_resolutions[sel].VFrontPorch;
	fb->hdmio_timing.VSyncWidth = vtc_resolutions[sel].VSyncWidth;
	fb->hdmio_timing.VBackPorch = vtc_resolutions[sel].VBackPorch;
	fb->hdmio_timing.VSyncPolarity = vtc_resolutions[sel].VSyncPolarity;

	status = fmc_imageon_hdmio_init(&(fb->fmc_imageon), 1,    // hdmioEnable = 1
			&(fb->hdmio_timing),  // pTiming
			1                     // waitHPD = 0
	);
	if (!status) // fmc_imageon_hdmio_init return 1 on success
		xil_printf("FAILED\n\r");
	else
		xil_printf("OK\n\r");

	/*
	 * initialize timing controller
	 */
	xil_printf("Initialize Timing Controller ... ");
	vtc->timing_sel = sel;

	XVtc_Disable(&vtc->Vtc);

	fill_vtc_timing(vtc);
	set_vtc_timing(vtc);

	xil_printf(" %s ", vtc_resolutions[sel].pName);

	xil_printf(" OK\n\r");

	return XST_SUCCESS;
}

#include "frame_size.h"
//#include "image_cores.h"

int main() {

	int status;

	u8 key;
	u8 menu_print_en = 1;
	double dummy;

	init_platform();

#ifdef XPAR_MICROBLAZE_USE_ICACHE
	Xil_ICacheEnable();
#endif
#ifdef XPAR_MICROBLAZE_USE_DCACHE
	Xil_DCacheEnable();
#endif

	xil_printf("\n\r\n\r");
	xil_printf("******************************\n\r");
	xil_printf("*  Signal Processing Dept.   *\n\r");
	xil_printf("*  UTIA AV CR, v.v.i.        *\n\r");
	xil_printf("*  Toshiba Sensor In IMAGEON *\n\r");
	xil_printf("*  [[Sobel edge detection]]  *\n\r");
	xil_printf("*  Full HD HDMI  Out IMAGEON *\n\r");
	xil_printf("*  C:/S/t20q1t1/hio          *\n\r");
	xil_printf("******************************\n\r");
	xil_printf("\n\r\n\r");


	/*
	 * HDMI output - IMAGEON on FMC2
	 */
	fb.uBaseAddr_IIC_FmcImageon = XPAR_FMC2_IMAGEON_IIC_BASEADDR;
	fb.uDeviceId_VTC_HdmiiDetector = 0; // no detector used in vdma branch
	fb.uDeviceId_VTC_HdmioGenerator = 0; // no generator needed in vdma branch
	fb.bVerbose = 0;

	/*
	 * Initialize I2C for FMC2
	 */
	xil_printf("IIC IMAGEON Initialization ... ");
	status = fmc_iic_xps_init(&(fb.fmc_imageon_iic),
			"\tHDMII-IMAGEON IIC Controller", fb.uBaseAddr_IIC_FmcImageon);
	if (!status) {
		xil_printf("FAILED\n\r");
		return XST_FAILURE;
	}

	status = fmc_imageon_init(&(fb.fmc_imageon), "\tHDMII-IMAGEON PC Input",
			&(fb.fmc_imageon_iic));
	if (!status) {  // fmc_imageon_hdmio_init return 1 on success
		xil_printf("FAILED\n\r");
		return XST_FAILURE;
	} else {
		xil_printf("OK\n\r");
	}

	fb.fmc_imageon.bVerbose = fb.bVerbose;

	/*
	 * Initialize IMAGEON HDMI Output
	 */
	xil_printf("HDMIO Initialization ... ");
	fb.hdmio_timing.IsHDMI = 0;

	fb.hdmio_timing.HActiveVideo =
			vtc_resolutions[VTC_1920x1080p60].HActiveVideo;
	fb.hdmio_timing.HFrontPorch = vtc_resolutions[VTC_1920x1080p60].HFrontPorch;
	fb.hdmio_timing.HSyncWidth = vtc_resolutions[VTC_1920x1080p60].HSyncWidth;
	fb.hdmio_timing.HBackPorch = vtc_resolutions[VTC_1920x1080p60].HBackPorch;
	fb.hdmio_timing.HSyncPolarity =
			vtc_resolutions[VTC_1920x1080p60].HSyncPolarity;

	fb.hdmio_timing.VActiveVideo =
			vtc_resolutions[VTC_1920x1080p60].VActiveVideo;
	fb.hdmio_timing.VFrontPorch = vtc_resolutions[VTC_1920x1080p60].VFrontPorch;
	fb.hdmio_timing.VSyncWidth = vtc_resolutions[VTC_1920x1080p60].VSyncWidth;
	fb.hdmio_timing.VBackPorch = vtc_resolutions[VTC_1920x1080p60].VBackPorch;
	fb.hdmio_timing.VSyncPolarity =
			vtc_resolutions[VTC_1920x1080p60].VSyncPolarity;

	status = fmc_imageon_hdmio_init(&(fb.fmc_imageon), 1,     // hdmioEnable = 1
			&(fb.hdmio_timing), // pTiming
			1                   // waitHPD = 0
	);
	if (!status) {  // fmc_imageon_hdmio_init return 1 on success
		xil_printf("FAILED\n\r");
		return XST_FAILURE;
	} else {
		xil_printf("OK\n\r");
	}

	/*
	 * initialize timing controller
	 */
	xil_printf("Initialize Timing Controller ... ");
	vtc.timing_sel = VTC_1920x1080p60;

	vtc.VtcCfgPtr = XVtc_LookupConfig(XPAR_VTC_0_DEVICE_ID);
	if (vtc.VtcCfgPtr == NULL) {
		xil_printf("Failed to get VTC config\n\r");
		return XST_FAILURE;
	}

	status = XVtc_CfgInitialize(&vtc.Vtc, vtc.VtcCfgPtr,
			vtc.VtcCfgPtr->BaseAddress);
	if (status != XST_SUCCESS) {
		xil_printf("Failed to init VTC");
		return XST_FAILURE;
	}

	XVtc_Disable(&vtc.Vtc);

	fill_vtc_timing(&vtc);
	set_vtc_timing(&vtc);

	xil_printf(" %s ", vtc_resolutions[VTC_1920x1080p60].pName);

	xil_printf(" OK\n\r");

	/*
	 * initialize VDMA output to display FB
	 */
	xil_printf("Initialize VDMA ... ");

	fb.uDeviceId_VDMA_HdmiFrameBuffer = XPAR_AXIVDMA_0_DEVICE_ID;
	fb.uBaseAddr_MEM_HdmiFrameBuffer = FB_BASE;        // frame storage for vdma
	fb.uNumFrames_HdmiFrameBuffer = XPAR_AXIVDMA_0_NUM_FSTORES;
	fb.vdma_tx_bytes = 2;
	fb.vdma_rx_bytes = 2;

	fb_store_res.x = MAX_RES_X;  // always set to maximum
	fb_store_res.y = MAX_RES_Y;  // always set to maximum
	fb_res.x = vtc_resolutions[VTC_1920x1080p60].HActiveVideo;
	fb_res.y = vtc_resolutions[VTC_1920x1080p60].VActiveVideo;

	xil_printf("Common Init ... ");
	status = vfb_common_init(fb.uDeviceId_VDMA_HdmiFrameBuffer, // uDeviceId
			&fb.vdma_hdmi                      // pAxiVdma
	);
	if (status != XST_SUCCESS) {
		xil_printf("FAILED\n\r");
		return XST_FAILURE;
	}

	xil_printf("TX Init ... ");
	status = vfb_tx_init(&fb.vdma_hdmi,                    // pAxiVdma
			&fb.vdmacfg_hdmi_read,            // pReadCfg
			&fb_res,                          // uVideoResolution
			&fb_store_res,       // uStorageResolution  -- always set to maximum
			fb.uBaseAddr_MEM_HdmiFrameBuffer, // uMemAddr
			fb.uNumFrames_HdmiFrameBuffer,    // uNumFrames
			&offset,                          // horizontal and vertical offset
			fb.vdma_tx_bytes                  // number of output bytes
	);
	if (status != XST_SUCCESS) {
		xil_printf("FAILED\n\r");
		return XST_FAILURE;
	} else {
		xil_printf("OK\n\r");
	}

	// initialize VDMA input
	xil_printf("RX Init ... ");
	status = vfb_rx_init( &(fb.vdma_hdmi),                   // pAxiVdma
			&(fb.vdmacfg_hdmi_write),          // pWriteCfg
			&fb_res,                           // uVideoResolution     -- use currently locked resolution
			&fb_store_res,                     // uStorageResolution   -- always set to maximum
			fb.uBaseAddr_MEM_HdmiFrameBuffer,  // uMemAddr
			fb.uNumFrames_HdmiFrameBuffer,     // uNumFrames
			fb.vdma_rx_bytes                   // number of bytes per pixel
	);
	if (status != XST_SUCCESS) {
		xil_printf("FAILED");
		return XST_FAILURE;
	} else {
		xil_printf("OK\n\r");
	}


	//***********************************************************************************

	while (1) {


		// initialize Camera
		xil_printf("TCM3232PB Init ... ");
		status = tcm_init( &fb );
		if ( status != XST_SUCCESS) {
			xil_printf("%d FAILED\n\r", status);
			return XST_FAILURE;
		}
		xil_printf("OK\n\r");


		/* print menu */
		//       print_menu(&menu_print_en);
		/*
		 * GUI - terminal
		 */
		key = 's';                  // ui_getch_nb();
		if (key != 0xff) {

			switch (key) {

			case 'm':
				menu_print_en = 1;
				break;

			case 'x':
				xil_printf("Exiting\n\r");
				return XST_FAILURE;
				break;

			case 's':
				xil_printf("Running sobel ....\r\n");
				key = ' ';

				// Status of AXI VDMA
				vfb_dump_registers(&fb.vdma_hdmi);
				if (vfb_check_errors(&fb.vdma_hdmi)) {
					vfb_dump_registers(&fb.vdma_hdmi);
				}

				/*
				 * Check available number of framebuffers
				 */
				if (fb.uNumFrames_HdmiFrameBuffer < 8) {
					xil_printf(
							"ERROR; not enough frame buffers to run sobel demo\r\n");
					break;
				}

				unsigned writingFrame = 0; // frame ktery se zapisuje z hdmi vstupu
				unsigned nextInputFrame = fb.uNumFrames_HdmiFrameBuffer - 1; // frame ktery je pripraven az se zacne zpracovavat
				unsigned inputFrame = fb.uNumFrames_HdmiFrameBuffer - 2; // frame ktery se prave zpracovava
				unsigned previousInputFrame = fb.uNumFrames_HdmiFrameBuffer - 3; // frame ktery se zpracovaval predtim
				unsigned outputFrame = fb.uNumFrames_HdmiFrameBuffer - 4; // frame pripraveny na dalsi vystup
				unsigned nextOutputFrame = fb.uNumFrames_HdmiFrameBuffer - 5; // frame do ktereho se zapisuje vysledek zpracovani
				unsigned readingFrame = fb.uNumFrames_HdmiFrameBuffer - 6; // frame ktery se cte do hdmi vystupu

				// setup parking
				XAxiVdma_StartParking(&fb.vdma_hdmi, writingFrame,
						XAXIVDMA_WRITE);

				XAxiVdma_StartParking(&fb.vdma_hdmi, readingFrame,
						XAXIVDMA_READ);

				xil_printf("Parking started\r\n");

				while (1) {

					while (writingFrame
							!= XAxiVdma_CurrFrameStore(&fb.vdma_hdmi,
									XAXIVDMA_WRITE)) {
					};


					while (readingFrame
							!= XAxiVdma_CurrFrameStore(&fb.vdma_hdmi,
									XAXIVDMA_READ)) {
					};

					/*
					 * send previous one and move parking to next
					 */
					advanceFStore(writingFrame);
					advanceFStore(nextInputFrame);
					advanceFStore(inputFrame);
					advanceFStore(previousInputFrame);
					advanceFStore(outputFrame);
					advanceFStore(nextOutputFrame);
					advanceFStore(readingFrame);

					/*
					 * Park to new positions
					 */
					XAxiVdma_StartParking(&fb.vdma_hdmi, readingFrame,
							XAXIVDMA_READ); // set parking to next, do not wait until switch
					XAxiVdma_StartParking(&fb.vdma_hdmi, writingFrame,
							XAXIVDMA_WRITE); // set parking to next, do not wait until switch

					/*
					 * Do all computations over inputFrame, previousInputFrame and outputFrame
					 */

					if (key == 'x')
						continue;  // exit sobel
					sobel_demo_processing(
							(unsigned short*) fb.vdmacfg_hdmi_read.FrameStoreStartAddr[inputFrame],
							(unsigned short*) fb.vdmacfg_hdmi_read.FrameStoreStartAddr[outputFrame]);
				}
				break;
			}
		}

	}

	Xil_DCacheDisable();
	Xil_ICacheDisable();

	cleanup_platform();

	return XST_SUCCESS;
}

